package org.jdtaus.core.io.util;

import java.util.HashMap;
import java.util.Map;
import java.util.Locale;
import java.util.ResourceBundle;
import java.text.MessageFormat;

abstract class CoalescingFileOperationsBundle{ 

/** <pre> A write operation bypassed the cache. Consider increasing the cache. Current block size is {0,number} and current cache size is {1,number}. {2,number} bytes could not be cached.</pre>. */
public static String getWriteBypassesCacheText(final Locale locale) {
    return getMessage("writeBypassesCache", locale);
}

/** <pre> A write operation bypassed the cache. Consider increasing the cache. Current block size is {0,number} and current cache size is {1,number}. {2,number} bytes could not be cached.</pre>. */
public static MessageFormat getWriteBypassesCacheMessage(Locale locale) {
    if(locale == null) { locale = Locale.getDefault(); }
return new MessageFormat(getMessage("writeBypassesCache", locale), locale);
}

/** <pre> A read operation bypassed the cache. Consider increasing the cache. Current block size is {0,number} and current cache size is {1,number}. {2,number} bytes could not be cached.</pre>. */
public static String getReadBypassesCacheText(final Locale locale) {
    return getMessage("readBypassesCache", locale);
}

/** <pre> A read operation bypassed the cache. Consider increasing the cache. Current block size is {0,number} and current cache size is {1,number}. {2,number} bytes could not be cached.</pre>. */
public static MessageFormat getReadBypassesCacheMessage(Locale locale) {
    if(locale == null) { locale = Locale.getDefault(); }
return new MessageFormat(getMessage("readBypassesCache", locale), locale);
}

/** <pre> Instance is closed - cannot perform I/O.</pre>. */
public static String getAlreadyClosedText(final Locale locale) {
    return getMessage("alreadyClosed", locale);
}

/** <pre> Instance is closed - cannot perform I/O.</pre>. */
public static MessageFormat getAlreadyClosedMessage(Locale locale) {
    if(locale == null) { locale = Locale.getDefault(); }
return new MessageFormat(getMessage("alreadyClosed", locale), locale);
}

private static final Map cache = new HashMap();

private static String getMessage(final String key, Locale locale) {
    if(locale == null) {locale = Locale.getDefault(); }
    Map msgCache = (Map) cache.get(locale);
    if(msgCache == null)
    {
        msgCache = new HashMap();
        cache.put(locale, msgCache);
    }

    String msg = (String) msgCache.get(key);
    if(msg == null)
    {
        msg = ResourceBundle.getBundle(
        "org.jdtaus.core.io.util.CoalescingFileOperations", locale).getString(key);
        msgCache.put(key, msg);
    };
    return msg;
}

}
